"""
The Dcmtk client uses dcmtk binaries (pre-requisite) along with files written and read by pydicom.
It also has a storage SCP that runs at all times, as opposed to the transient listeners
spawned by PynetDicomClient.

DCMDICTPATH and (depending on the installation) SCPCFGPATH envrionment variables are
required.
"""
import logging
import os
import re
import subprocess
from subprocess import PIPE
import shutil
import tempfile
import threading
import glob
from collections import defaultdict

from typing import List, Optional, Iterable, Tuple

import pydicom
from pydicom import dcmread
from pydicom.dataset import Dataset

from .base_client import BaseDicomClient
from .utils import process_and_write_png_from_file, copy_dicom_attributes, \
    set_undefined_tags_to_blank, dicom_filename

logger = logging.getLogger(__name__)

# http://dicom.nema.org/medical/dicom/current/output/html/part07.html#chapter_C
status_success_or_pending = [0x0000, 0xFF00, 0xFF01]

dcmtk_error_codes = {
    # No data (timeout in non-blocking mode)
    'dcmnet-DIMSEC_NODATAAVAILABLE': (6, 0x207)
}
"""
This is a partial list of DCMTK error code constants, separated by module.

Codes are printed as a pair of module code + error code, e.g. `0006:0207`

Module codes are defined in `dcerror.h`.
https://github.com/DCMTK/dcmtk/blob/31ae87d57edf3f5a441cae64869c7337b29213b2/dcmdata/include/dcmtk/dcmdata/dcerror.h#L36-L75

Error codes are defined in multiple places within the DCMTK source, such as
https://github.com/DCMTK/dcmtk/blob/master/dcmnet/libsrc/cond.cc for networking errors.
"""

backoff_padding = 20
"""
If retrying timeouts with a backoff is on, this padding will be added to the existing timeout before retrying
"""

move_lock = threading.Lock()


class DcmtkDicomClient(BaseDicomClient):
    def __init__(
        self,
        client_ae,
        remote_ae,
        pacs_url,
        pacs_port,
        dicom_dir,
        dcmtk_profile: str = "AllDICOM",
        timeout=20,
        storescp_extra_args=None,
        movescu_extra_args=None,
        findscu_extra_args=None,
        retry_timeouts_with_backoff=False,
        *args, **kwargs,
    ):
        """
        :param client_ae: Name for this client Association Entity. {client_ae}:11113
            needs to be registered with the remote PACS in order for C-MOVE to work
        :param pacs_url: Remote PACS URL
        :param pacs_port: Remote PACS port (usually 11112)
        :param dicom_dir: Root dir for storage of *.dcm files.
        :param dcmtk_profile: Profile name from storescp.cfg to use
        :param timeout: Connection and DICOM timeout in seconds
        :param storescp_extra_args: Optional array of extra arguments to supply to the `storescp` invocation
        :param findscu_extra_args: Optional array of extra arguments to supply to the `findscu` invocation
        :param movescu_extra_args: Optional array of extra arguments to supply to the `movescu` invocation
        :param retry_timeouts_with_backoff: If true, will retry failures due to timeout, with a longer timeout period.
            default=False

        Note: the `dcmtk_profile` variable refers to the profile name defined
        in the `storescp.cfg` configuration file, the location of which is
        indicated by the `SCPCFGPATH` environment variable. This configuration
        tells the `storescp` which presentation contexts to accept when
        negotiating an association with a storage SCU.

        "AllDICOM" is a profile name that exists in the default storescp
        configuration file which accepts almost all presentation contexts. If a
        more restricted set of contexts is desired, the configuration file
        should be updated and a new profile name should be passed in as an
        argument.

        For the use of `*_extra_args`: the array of arguments can be easily
        generated from a plain string with `shlex.split()`, e.g.

        >>> import shlex
        >>> extra_arguments = '--some-arg value --complex-string "multiple words here"'
        >>> storescp_extra_args = shlex.split(extra_arguments)
        >>> storescp_extra_args
        ['--some-arg', 'value', '--complex-string', 'multiple words here']
        """
        self.client_ae = client_ae
        self.remote_ae = remote_ae
        self.pacs_url = pacs_url
        self.pacs_port = str(pacs_port)
        self.dicom_dir = dicom_dir
        self.dicom_tmp_dir = os.path.join(self.dicom_dir, 'tmp')
        self.timeout = timeout
        self.listener_port = str(11113)
        self.storescp_extra_args = storescp_extra_args or []
        self.findscu_extra_args = findscu_extra_args or []
        self.movescu_extra_args = movescu_extra_args or []
        self.retry_timeouts_with_backoff = retry_timeouts_with_backoff
        self.dcmtk_profile = dcmtk_profile
        if logger.getEffectiveLevel() <= logging.DEBUG:
            self.logger_args = ['-v', '-d']
        else:
            self.logger_args = []

        # ensure binaries are available
        subprocess.run(['storescp', '--version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)
        subprocess.run(['movescu', '--version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)
        subprocess.run(['findscu', '--version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)

        # run 1 storescp listener at all times
        os.makedirs(self.dicom_tmp_dir, exist_ok=True)
        dcm_dict_dir = os.path.dirname(os.environ['DCMDICTPATH'])
        if 'SCPCFGPATH' in os.environ:
            storescp_config_path = os.environ['SCPCFGPATH']
        else:
            # fallback path typical to some dcmtk installations
            storescp_config_path = os.path.join(dcm_dict_dir, '../../etc/dcmtk/storescp.cfg')

        # TODO storescp logging is going to stdout: should have self.logger redirect
        storescp_args = ['storescp', '--fork', '--aetitle', client_ae,
                         *self.logger_args,
                         '--output-directory', self.dicom_tmp_dir,
                         '--filename-extension', '.dcm',
                         '--config-file', storescp_config_path, self.dcmtk_profile,
                         *self.storescp_extra_args,
                         self.listener_port]
        self.process = subprocess.Popen(storescp_args)

    def verify(self) -> bool:
        echoscu_args = ['echoscu', '--aetitle', self.remote_ae, '--call', self.client_ae,
                        *self._get_timeout_args(), self.pacs_url, self.pacs_port, *self.logger_args]

        result = subprocess.run(echoscu_args, stdout=PIPE, stderr=PIPE, universal_newlines=True)

        logger.debug(result.args)
        logger.debug(result.stdout)
        logger.debug(result.stderr)

        return result.returncode == 0

    def _get_timeout_args(self, is_retry=False):
        offset = backoff_padding if is_retry else 0
        return ['--timeout', str(self.timeout + offset),
                '--dimse-timeout', str(self.timeout + offset)]

    def _get_study_search_dataset(self, study_date_tag=None):
        search_dataset = Dataset()
        search_dataset.PatientID = None
        search_dataset.PatientName = ''
        search_dataset.PatientBirthDate = None
        if study_date_tag is not None:
            search_dataset.StudyDate = study_date_tag
        else:
            search_dataset.StudyDate = ''
        search_dataset.StudyInstanceUID = ''
        search_dataset.QueryRetrieveLevel = 'STUDY'
        return search_dataset

    def _send_c_find(self, search_dataset, is_retry=False):
        result_datasets = []

        search_dataset.is_little_endian = True
        search_dataset.is_implicit_VR = True
        with tempfile.TemporaryDirectory() as tmpdirname:
            find_dataset_path = os.path.join(tmpdirname, 'find_input.dcm')
            pydicom.dcmwrite(find_dataset_path, search_dataset)

            output_dir = os.path.join(tmpdirname, 'find_output')
            os.mkdir(output_dir)

            findscu_args = ['findscu', '--aetitle', self.client_ae, *self.logger_args,
                            '--call', self.remote_ae,
                            *self._get_timeout_args(is_retry), '-S',
                            '-X', '--output-directory', output_dir, *self.findscu_extra_args,
                            self.pacs_url, self.pacs_port, find_dataset_path]
            result = subprocess.run(findscu_args, stdout=PIPE, stderr=PIPE, universal_newlines=True)
            logger.debug(result.args)
            logger.debug(result.stdout)
            logger.debug(result.stderr)

            if result.returncode != 0:
                logger.error(
                    f'C-FIND failure for search dataset: rc {result.returncode}')
                logger.error(search_dataset)
                return []

            if _check_dcmtk_message_for_timeout(result.stdout or result.stderr):
                if self.retry_timeouts_with_backoff and not is_retry:
                    logger.warning('C-FIND timed out, but retry is on. Trying again.')
                    return self._send_c_find(search_dataset, True)
                logger.error('C-FIND failure for search dataset: Timed out.')
                logger.error(search_dataset)
                return []

            for dcm_file in glob.glob(f'{output_dir}/*.dcm'):
                result_datasets.append(dcmread(dcm_file))

        return result_datasets

    def _send_c_move(self, move_dataset, output_dir, is_retry=False):
        if self.process.returncode is not None:
            msg = 'dcmrecv is not running, rc {self.process.returncode}'
            logger.error(msg)
            raise Exception(msg)

        with tempfile.TemporaryDirectory() as tmpdirname:
            move_dataset_path = os.path.join(tmpdirname, 'move_dataset.dcm')

            os.makedirs(output_dir, exist_ok=True)
            move_dataset.is_little_endian = True
            move_dataset.is_implicit_VR = True
            pydicom.dcmwrite(move_dataset_path, move_dataset)

            # even though storescp has `--fork`, the move lock is needed to tell datasets
            #  apart in the `dicom_tmp_dir`
            with move_lock:
                movescu_args = ['movescu', '--aetitle', self.client_ae, '--call',
                                self.remote_ae,
                                '--move', self.client_ae, '-S',  # study query level
                                *self._get_timeout_args(), *self.logger_args, *self.movescu_extra_args,
                                self.pacs_url, self.pacs_port, move_dataset_path]
                result = subprocess.run(movescu_args, stdout=PIPE, stderr=PIPE, universal_newlines=True)

                logger.debug(result.args)
                logger.debug(result.stdout)
                logger.debug(result.stderr)

                for result_item in os.listdir(self.dicom_tmp_dir):
                    # fully specify move destination to allow overwrites
                    shutil.move(os.path.join(self.dicom_tmp_dir, result_item),
                                os.path.join(output_dir, result_item))

            if result.returncode != 0:
                logger.error(f'C-MOVE failure for query: rc {result.returncode}')
                return False

            if _check_dcmtk_message_for_timeout(result.stdout or result.stderr):
                if self.retry_timeouts_with_backoff and not is_retry:
                    logger.warning('C-MOVE timed out, but retry is on. Trying again.')
                    return self._send_c_move(move_dataset, output_dir, True)
                logger.error('C-MOVE failure for search dataset: Timed out.')
                return False

            return True

    def search_patients(self, search_query: Optional[str] = None,
                        search_query_type: Optional[str] = None,
                        additional_tags: Optional[List[str]] = None,
                        wildcard: bool = True) -> List[Dataset]:
        '''
        :param search_query: String containing query value to c_find.
            PatientID, PatientName, or, if wildcard is True, any partial there of.
        :param search_query_type: Optional string that restricts patient search to 'PatientID' or 'PatientName'.
            If None, perform a C-FIND for the query once on PatientID and again on PatientName.
        :param additional_tags: Additional dicom attributes that should be
            included in the returned patient dicom dataset values object.
        :param wildcard: Boolean stating whether to search based on
            any PatientName or PatientID partial string (i.e. Sam would find Samuel).
        :returns: List of DICOM query responses for each patient matching the query.
        '''
        if wildcard:
            search_query = f'*{search_query}*'
        patient_id_to_datasets = defaultdict(Dataset)

        search_dataset = self._get_study_search_dataset()
        if search_query_type == 'PatientID' or search_query_type is None:
            search_dataset.PatientID = search_query
            self._search_patient_with_dataset(search_dataset, patient_id_to_datasets, additional_tags)
        if search_query_type == 'PatientName' or search_query_type is None:
            search_dataset.PatientName = search_query
            self._search_patient_with_dataset(search_dataset, patient_id_to_datasets, additional_tags)

        return list(patient_id_to_datasets.values())

    def _search_patient_with_dataset(self, search_dataset: Dataset,
                                     patient_id_to_datasets: defaultdict,
                                     additional_tags: Optional[List[str]] = None):
        '''
        This function does not return any values but rather modifies the patient_id_to_datasets argument in-place.
        '''
        set_undefined_tags_to_blank(search_dataset, additional_tags)
        responses = self._send_c_find(search_dataset)
        for study in responses:
            if hasattr(study, 'PatientID'):
                result = patient_id_to_datasets[study.PatientID]
                self.update_patient_result(result, study, additional_tags)

    def studies_for_patient(self, patient_id, study_date_tag=None, additional_tags=None) -> List[Dataset]:
        search_dataset = self._get_study_search_dataset(study_date_tag)
        search_dataset.PatientID = patient_id
        set_undefined_tags_to_blank(search_dataset, additional_tags)

        responses = self._send_c_find(search_dataset)
        datasets = []
        for dataset in responses:
            # Some PACS send back empty "Success" responses at the end of the list
            if hasattr(dataset, 'PatientID'):
                datasets.append(dataset)

        return datasets

    def search_series(self, query_dataset, additional_tags=None) -> List[Dataset]:
        additional_tags = additional_tags or []
        query_dataset.QueryRetrieveLevel = 'SERIES'
        additional_tags += [
            'Modality',
            'SeriesDescription',
            'SeriesDate',
            'SeriesTime',
        ]
        set_undefined_tags_to_blank(query_dataset, additional_tags)

        datasets = []
        responses = self._send_c_find(query_dataset)
        for series in responses:
            if hasattr(series, 'SeriesInstanceUID'):
                datasets.append(series)
        return datasets

    def series_for_study(self, study_id, modality_filter=None, additional_tags=None,
                         manual_count=True) -> \
            List[Dataset]:
        additional_tags = additional_tags or []

        dataset = Dataset()
        dataset.StudyInstanceUID = study_id
        dataset.QueryRetrieveLevel = 'SERIES'

        additional_tags += [
            'SeriesInstanceUID',
            'BodyPartExamined',
            'SeriesDescription',
            'SeriesDate',
            'SeriesTime',
            'StudyDate',
            'StudyTime',
            'NumberOfSeriesRelatedInstances',
        ]
        set_undefined_tags_to_blank(dataset, additional_tags)
        # TODO modality filtering not implemented
        dataset.Modality = ''

        raw_series_datasets = self._send_c_find(dataset)

        series_datasets = []
        for series in raw_series_datasets:
            valid_dicom = hasattr(series, 'SeriesInstanceUID')
            modality = getattr(series, 'Modality', '')
            match = modality_filter is None or modality in modality_filter
            if valid_dicom and match:
                ds = Dataset()
                ds.SeriesDescription = getattr(series, 'SeriesDescription', '')
                ds.BodyPartExamined = getattr(series, 'BodyPartExamined', None)
                ds.SeriesInstanceUID = series.SeriesInstanceUID
                ds.Modality = series.Modality
                copy_dicom_attributes(ds, series, additional_tags)
                ds.NumberOfSeriesRelatedInstances = self._determine_number_of_images(
                    series, manual_count)
                series_datasets.append(ds)

        return series_datasets

    def _determine_number_of_images(self, series, manual_count):
        answer_from_instance_count = series.NumberOfSeriesRelatedInstances
        if answer_from_instance_count:
            return answer_from_instance_count
        elif manual_count:
            return str(self._count_images_via_query(series))
        else:
            return None

    def _count_images_via_query(self, series):
        series_dataset = Dataset()
        series_dataset.SeriesInstanceUID = series.SeriesInstanceUID
        series_dataset.QueryRetrieveLevel = 'IMAGE'
        series_dataset.SOPInstanceUID = ''

        series_responses = self._send_c_find(series_dataset)
        image_count = 0
        for instance in series_responses:
            if hasattr(instance, 'SOPInstanceUID'):
                image_count += 1
        return image_count

    def images_for_series(self, study_id, series_id, additional_tags=None,
                          max_count=None) -> List[Dataset]:

        image_datasets = []

        series_dataset = Dataset()
        series_dataset.StudyInstanceUID = study_id
        series_dataset.SeriesInstanceUID = series_id
        series_dataset.QueryRetrieveLevel = 'IMAGE'
        series_dataset.SOPInstanceUID = ''
        set_undefined_tags_to_blank(series_dataset, additional_tags)

        series_responses = self._send_c_find(series_dataset)
        for instance in series_responses:
            if hasattr(instance, 'SOPInstanceUID'):
                ds = Dataset()
                ds.SeriesInstanceUID = instance.SeriesInstanceUID
                ds.SOPInstanceUID = instance.SOPInstanceUID
                copy_dicom_attributes(ds, instance, additional_tags)
                image_datasets.append(ds)
                if max_count and len(image_datasets) >= max_count:
                    break
        return image_datasets

    def fetch_images_as_dicom_files(self, study_id: str, series_id: str) -> Optional[str]:
        series_path = os.path.join(self.dicom_dir, series_id)

        dataset = Dataset()
        dataset.SeriesInstanceUID = series_id
        dataset.StudyInstanceUID = study_id
        dataset.QueryRetrieveLevel = 'SERIES'
        dataset.SOPInstanceUID = ''

        success = self._send_c_move(dataset, series_path)

        return series_path if success and os.path.exists(series_path) else None

    def fetch_image_as_dicom_file(self, study_id: str, series_id: str,
                                  sop_instance_id: str) -> Optional[str]:
        series_path = os.path.join(self.dicom_dir, series_id)
        dataset = Dataset()
        dataset.SeriesInstanceUID = series_id
        dataset.StudyInstanceUID = study_id
        dataset.SOPInstanceUID = sop_instance_id
        dataset.QueryRetrieveLevel = 'IMAGE'

        success = self._send_c_move(dataset, self.series_path)
        filepath = os.path.join(series_path, dicom_filename(dataset))

        return filepath if success and os.path.exists(filepath) else None

    def fetch_thumbnail(self, study_id: str, series_id: str) -> Optional[str]:
        # search for image IDs in the series
        find_dataset = Dataset()
        find_dataset.StudyInstanceUID = study_id
        find_dataset.SeriesInstanceUID = series_id
        find_dataset.QueryRetrieveLevel = 'IMAGE'
        find_dataset.SOPInstanceUID = ''
        image_responses = self._send_c_find(find_dataset)

        image_ids = []
        for dataset in image_responses:
            if hasattr(dataset, 'SOPInstanceUID'):
                image_ids.append(dataset.SOPInstanceUID)

        if not image_ids:
            logger.error(f'Failed to find any image instances for series {series_id}')
            return None

        # try to get the middle image in the series for the thumbnail:
        #  instance ID order is usually the same as slice order but not guaranteed
        #  by the standard.
        middle_image_id = image_ids[len(image_ids) // 2]
        return self._fetch_individual_slice_thumbnail(study_id, series_id, middle_image_id)

    def fetch_slice_thumbnail(self, study_id: str, series_id: str,
                              instance_id: str) -> Optional[str]:
        return self._fetch_individual_slice_thumbnail(study_id, series_id, instance_id)

    def _fetch_individual_slice_thumbnail(self, study_id: str, series_id: str,
                                          instance_id: str) -> Optional[str]:
        move_dataset = Dataset()
        move_dataset.StudyInstanceUID = study_id
        move_dataset.SeriesInstanceUID = series_id
        move_dataset.SOPInstanceUID = instance_id
        move_dataset.QueryRetrieveLevel = 'IMAGE'

        success = self._send_c_move(move_dataset, self.dicom_dir)

        # dcmtk puts modality prefixes in front of the instance IDs
        dcm_paths = glob.glob(os.path.join(self.dicom_dir, f'*{instance_id}.dcm'))
        if not success or not dcm_paths:
            logger.error(f'Failure to get thumbnail for {instance_id}')
            return None
        if len(dcm_paths) > 1:
            logger.error(f'Found duplicate thumbnails for {instance_id}: {dcm_paths}')
            return None

        dcm_path = dcm_paths[0]
        png_path = process_and_write_png_from_file(dcm_path)
        return png_path

    def send_datasets(self, datasets: Iterable[Dataset], override_remote_ae: str = None,
                      override_pacs_url: str = None, override_pacs_port: int = None) -> None:

        if override_remote_ae is not None and override_pacs_url is not None and override_pacs_port is not None:
            send_remote_ae = override_remote_ae
            send_port = str(override_pacs_port)
            send_url = override_pacs_url
        else:
            send_remote_ae = self.remote_ae
            send_port = self.pacs_port
            send_url = self.pacs_url

        for dataset in datasets:
            logger.info('Sending %s', dataset.SeriesInstanceUID)
            with tempfile.TemporaryDirectory() as tmpdirname:
                store_dcm_file = os.path.join(tmpdirname, 'store_dataset.dcm')
                pydicom.dcmwrite(store_dcm_file, dataset)
                storescu_args = ['storescu', '--aetitle', self.client_ae,
                                 '--call', send_remote_ae,
                                 *self._get_timeout_args(), *self.logger_args,
                                 send_url, send_port,
                                 store_dcm_file]

                result = subprocess.run(storescu_args, stdout=PIPE, stderr=PIPE,
                                        universal_newlines=True)
                logger.debug(result.args)
                logger.debug(result.stdout)
                logger.debug(result.stderr)
                if result.returncode != 0:
                    msg = f'Failure to send dataset with {dataset.SeriesInstanceUID}, rc {result.returncode}'
                    logger.error(msg)
                    raise Exception(msg)


def _check_dcmtk_message_for_error(dcmtk_message: str) -> Optional[Tuple[int, int]]:
    """
    This checks a message from DCMTK for a known error message pattern.

    :param dcmtk_message: The message to parse / check for an error message.
    :return: Either `None` if no error is detected, else a tuple with (module_code, err_code)

    This is useful, because for certain operations, DCMTK is likely to always return a zero exit code, even in cases
    of (nested) failure.

    For example, `findscu` calls will seemingly only return a non-zero exit code if the failure happens at the
    association level (e.g. on setup or abort).

    This is a known issue: https://support.dcmtk.org/redmine/issues/929
    """
    # Standard pattern is E: 0001:0002 ERROR MESSAGE
    pattern = re.compile(r"[EF]: ([\da-f]{4}:[\da-f]{4}) [^#\r\n]+$", flags=re.MULTILINE)

    # Only check last three lines, and in reverse order (last first)
    message_lines = dcmtk_message.splitlines()[-3:]
    message_lines.reverse()
    for line in message_lines:
        match = pattern.search(line)
        if match:
            return tuple(map(lambda code: int(code, 16), match.group(1).split(':')))

    return None


def _check_dcmtk_message_for_timeout(dcmtk_message: str) -> bool:
    error_tuple = _check_dcmtk_message_for_error(dcmtk_message)
    return error_tuple == dcmtk_error_codes['dcmnet-DIMSEC_NODATAAVAILABLE']



import logging
import os
import threading
from contextlib import contextmanager
from collections import defaultdict
from typing import List, Optional, Iterable

from pydicom.dataset import Dataset, FileDataset
from pynetdicom import AE, StoragePresentationContexts, evt
from pynetdicom import PYNETDICOM_IMPLEMENTATION_UID, PYNETDICOM_IMPLEMENTATION_VERSION
from pynetdicom.sop_class import Verification, \
    StudyRootQueryRetrieveInformationModelFind, StudyRootQueryRetrieveInformationModelMove

from .base_client import BaseDicomClient
from .utils import process_and_write_png_from_file, copy_dicom_attributes,\
    set_undefined_tags_to_blank, dicom_filename

logger = logging.getLogger(__name__)

# http://dicom.nema.org/medical/dicom/current/output/html/part07.html#chapter_C
status_success_or_pending = [0x0000, 0xFF00, 0xFF01]

C_FIND_QUERY_MODEL = StudyRootQueryRetrieveInformationModelFind
C_MOVE_QUERY_MODEL = StudyRootQueryRetrieveInformationModelMove


class PynetDicomClient(BaseDicomClient):
    def __init__(self, client_ae, remote_ae, pacs_url, pacs_port, dicom_dir, timeout=5,
                 *args, **kwargs):
        """
        :param client_ae: Name for this client Association Entity. {client_ae}:11113
            needs to be registered with the remote PACS in order for C-MOVE to work
        :param pacs_url: Remote PACS URL
        :param pacs_port: Remote PACS port (usually 11112)
        :param dicom_dir: Root dir for storage of *.dcm files.
        :param timeout: Connection and DICOM timeout in seconds
        """
        self.client_ae = client_ae
        self.remote_ae = remote_ae
        self.pacs_url = pacs_url
        self.pacs_port = pacs_port
        self.dicom_dir = dicom_dir
        self.timeout = timeout

    def verify(self) -> bool:
        ae = AE(ae_title=self.client_ae)
        ae.add_requested_context(Verification)
        # setting timeout here doesn't appear to have any effect
        ae.network_timeout = self.timeout

        with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as assoc:
            logger.debug('Association accepted by the peer')
            # Send a DIMSE C-ECHO request to the peer
            # status is a pydicom Dataset object with (at a minimum) a
            # (0000, 0900) Status element
            status = assoc.send_c_echo()

            # Output the response from the peer
            if status.Status in status_success_or_pending:
                logger.debug('C-ECHO Response: 0x{0:04x}'.format(status.Status))
                return True
            else:
                logger.warning('C-ECHO Failure Response: 0x{0:04x}'.format(status.Status))
                return False

        return False

    def search_patients(self, search_query: str, additional_tags: List[str] = None,
                        wildcard: bool = None) -> List[Dataset]:
        ae = AE(ae_title=self.client_ae)
        ae.add_requested_context(StudyRootQueryRetrieveInformationModelFind)

        if wildcard:
            search_query = f'*{search_query}*'
        patient_id_to_datasets = defaultdict(Dataset)

        with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as assoc:
            id_responses = _find_patients(assoc, 'PatientID', search_query, additional_tags)
            for study in checked_responses(id_responses):
                if hasattr(study, 'PatientID'):
                    result = patient_id_to_datasets[study.PatientID]
                    self.update_patient_result(result, study, additional_tags)

        # consecutive find must be in separate associations
        with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as assoc:
            name_responses = _find_patients(assoc, 'PatientName', search_query, additional_tags)
            for study in checked_responses(name_responses):
                if hasattr(study, 'PatientID'):
                    result = patient_id_to_datasets[study.PatientID]
                    self.update_patient_result(result, study, additional_tags)

        return list(patient_id_to_datasets.values())

    def studies_for_patient(self, patient_id, study_date_tag=None, additional_tags=None) -> List[Dataset]:
        ae = AE(ae_title=self.client_ae)
        ae.add_requested_context(StudyRootQueryRetrieveInformationModelFind)

        with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as assoc:
            responses = _find_patients(assoc, 'PatientID', f'{patient_id}', study_date_tag, additional_tags)

            datasets = []
            for dataset in checked_responses(responses):
                # Some PACS send back empty "Success" responses at the end of the list
                if hasattr(dataset, 'PatientID'):
                    datasets.append(dataset)

            return datasets

    def search_series(self, query_dataset, additional_tags=None) -> List[Dataset]:
        additional_tags = additional_tags or []
        query_dataset.QueryRetrieveLevel = 'SERIES'
        additional_tags += [
            'Modality',
            'SeriesDescription',
            'SeriesDate',
            'SeriesTime',
        ]
        set_undefined_tags_to_blank(query_dataset, additional_tags)
        ae = AE(ae_title=self.client_ae)
        ae.add_requested_context(StudyRootQueryRetrieveInformationModelFind)

        datasets = []
        with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as assoc:
            responses = assoc.send_c_find(query_dataset, query_model=C_FIND_QUERY_MODEL)
            for series in checked_responses(responses):
                if hasattr(series, 'SeriesInstanceUID'):
                    datasets.append(series)
        return datasets

    def series_for_study(self, study_id, modality_filter=None, additional_tags=None,
                         manual_count=True) -> List[Dataset]:
        additional_tags = additional_tags or []
        ae = AE(ae_title=self.client_ae)
        ae.add_requested_context(StudyRootQueryRetrieveInformationModelFind)

        with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as assoc:
            dataset = Dataset()
            dataset.StudyInstanceUID = study_id
            dataset.QueryRetrieveLevel = 'SERIES'

            additional_tags += [
                'SeriesInstanceUID',
                'BodyPartExamined',
                'SeriesDescription',
                'SeriesDate',
                'SeriesTime',
                'StudyDate',
                'StudyTime',
                'NumberOfSeriesRelatedInstances',
            ]
            set_undefined_tags_to_blank(dataset, additional_tags)
            # Filtering modality with 'MR\\CT' doesn't seem to work with pynetdicom
            dataset.Modality = ''

            responses = assoc.send_c_find(dataset, query_model=C_FIND_QUERY_MODEL)

            series_datasets = []
            for series in checked_responses(responses):
                valid_dicom = hasattr(series, 'SeriesInstanceUID')
                modality = getattr(series, 'Modality', '')
                match = modality_filter is None or modality in modality_filter
                if valid_dicom and match:
                    ds = Dataset()
                    ds.SeriesDescription = getattr(series, 'SeriesDescription', '')
                    ds.BodyPartExamined = getattr(series, 'BodyPartExamined', None)
                    ds.SeriesInstanceUID = series.SeriesInstanceUID
                    ds.Modality = series.Modality
                    ds.SeriesDate = series.SeriesDate
                    ds.SeriesTime = series.SeriesTime
                    copy_dicom_attributes(ds, series, additional_tags)
                    instance_count = self._determine_number_of_images(ae, series, manual_count)
                    ds.NumberOfSeriesRelatedInstances = instance_count
                    series_datasets.append(ds)

        return series_datasets

    def _determine_number_of_images(self, ae, series, manual_count):
        answer_from_instance_count = series.NumberOfSeriesRelatedInstances
        if answer_from_instance_count:
            return answer_from_instance_count
        elif manual_count:
            return str(self._count_images_via_query(ae, series))
        else:
            return None

    def _count_images_via_query(self, ae, series):
        with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as series_assoc:
            series_dataset = Dataset()
            series_dataset.SeriesInstanceUID = series.SeriesInstanceUID
            series_dataset.QueryRetrieveLevel = 'IMAGE'
            series_dataset.SOPInstanceUID = ''

            series_responses = series_assoc.send_c_find(series_dataset, query_model=C_FIND_QUERY_MODEL)
            image_count = 0
            for instance in checked_responses(series_responses):
                if hasattr(instance, 'SOPInstanceUID'):
                    image_count += 1
        return image_count

    def images_for_series(self, study_id, series_id, additional_tags=None, max_count=None) -> List[Dataset]:

        ae = AE(ae_title=self.client_ae)
        ae.add_requested_context(StudyRootQueryRetrieveInformationModelFind)

        image_datasets = []
        with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as series_assoc:
            series_dataset = Dataset()
            series_dataset.StudyInstanceUID = study_id
            series_dataset.SeriesInstanceUID = series_id
            series_dataset.QueryRetrieveLevel = 'IMAGE'
            series_dataset.SOPInstanceUID = ''
            set_undefined_tags_to_blank(series_dataset, additional_tags)

            series_responses = series_assoc.send_c_find(series_dataset, query_model=C_FIND_QUERY_MODEL)
            for instance in checked_responses(series_responses):
                if hasattr(instance, 'SOPInstanceUID'):
                    ds = Dataset()
                    ds.SeriesInstanceUID = instance.SeriesInstanceUID
                    ds.SOPInstanceUID = instance.SOPInstanceUID
                    copy_dicom_attributes(ds, instance, additional_tags)
                    image_datasets.append(ds)
                    if max_count and len(image_datasets) >= max_count:
                        break
        return image_datasets

    def fetch_images_as_dicom_files(self, study_id: str, series_id: str) -> Optional[str]:

        series_path = os.path.join(self.dicom_dir, series_id)
        with storage_scp(self.client_ae, series_path) as scp:
            ae = AE(ae_title=self.client_ae)
            ae.add_requested_context(StudyRootQueryRetrieveInformationModelMove)

            with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as assoc:
                dataset = Dataset()
                dataset.SeriesInstanceUID = series_id
                dataset.StudyInstanceUID = study_id
                dataset.QueryRetrieveLevel = 'SERIES'
                dataset.SOPInstanceUID = ''

                if scp.is_alive():
                    responses = assoc.send_c_move(dataset, scp.ae_title, query_model=C_MOVE_QUERY_MODEL)
                else:
                    raise Exception(f'Storage SCP failed to start for series {series_id}')

                check_responses(responses)
                return series_path if os.path.exists(series_path) else None

    def fetch_image_as_dicom_file(self, study_id: str, series_id: str, sop_instance_id: str) -> Optional[str]:
        series_path = os.path.join(self.dicom_dir, series_id)
        with storage_scp(self.client_ae, series_path) as scp:
            ae = AE(ae_title=self.client_ae)
            ae.add_requested_context(StudyRootQueryRetrieveInformationModelMove)

            with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as assoc:
                dataset = Dataset()
                dataset.SeriesInstanceUID = series_id
                dataset.StudyInstanceUID = study_id
                dataset.SOPInstanceUID = sop_instance_id
                dataset.QueryRetrieveLevel = 'IMAGE'

                if scp.is_alive():
                    responses = assoc.send_c_move(dataset, scp.ae_title, query_model=C_MOVE_QUERY_MODEL)
                else:
                    raise Exception(f'Storage SCP failed to start for series {series_id}')

                check_responses(responses)
                filepath = scp.path_for_dataset_instance(dataset)
                return filepath if os.path.exists(filepath) else None
        return None

    def fetch_thumbnail(self, study_id: str, series_id: str) -> Optional[str]:
        ae = AE(ae_title=self.client_ae)
        ae.add_requested_context(StudyRootQueryRetrieveInformationModelFind)
        ae.add_requested_context(StudyRootQueryRetrieveInformationModelMove)

        with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as assoc:
            # search for image IDs in the series
            find_dataset = Dataset()
            find_dataset.StudyInstanceUID = study_id
            find_dataset.SeriesInstanceUID = series_id
            find_dataset.QueryRetrieveLevel = 'IMAGE'
            find_dataset.SOPInstanceUID = ''
            find_response = assoc.send_c_find(find_dataset, query_model=C_FIND_QUERY_MODEL)

            image_ids = []
            for dataset in checked_responses(find_response):
                if hasattr(dataset, 'SOPInstanceUID'):
                    image_ids.append(dataset.SOPInstanceUID)

            if not image_ids:
                return None

            with storage_scp(self.client_ae, self.dicom_dir) as scp:
                # try to get the middle image in the series for the thumbnail:
                #  instance ID order is usually the same as slice order but not guaranteed
                #  by the standard.
                middle_image_id = image_ids[len(image_ids) // 2]
                move_dataset = Dataset()
                move_dataset.StudyInstanceUID = study_id
                move_dataset.SeriesInstanceUID = series_id
                move_dataset.SOPInstanceUID = middle_image_id
                move_dataset.QueryRetrieveLevel = 'IMAGE'

                if scp.is_alive():
                    move_responses = assoc.send_c_move(move_dataset, scp.ae_title,
                                                       query_model=C_MOVE_QUERY_MODEL)
                else:
                    raise Exception(f'Storage SCP failed to start for series {series_id}')

                check_responses(move_responses)

        dcm_path = os.path.join(self.dicom_dir, f'{middle_image_id}.dcm')
        png_path = process_and_write_png_from_file(dcm_path)
        return png_path

    def fetch_slice_thumbnail(self, study_id: str, series_id: str,
                              instance_id: str) -> Optional[str]:
        ae = AE(ae_title=self.client_ae)
        ae.add_requested_context(StudyRootQueryRetrieveInformationModelFind)
        ae.add_requested_context(StudyRootQueryRetrieveInformationModelMove)

        with association(ae, self.pacs_url, self.pacs_port, self.remote_ae) as assoc:
            with storage_scp(self.client_ae, self.dicom_dir) as scp:
                move_dataset = Dataset()
                move_dataset.StudyInstanceUID = study_id
                move_dataset.SeriesInstanceUID = series_id
                move_dataset.SOPInstanceUID = instance_id
                move_dataset.QueryRetrieveLevel = 'IMAGE'

                if scp.is_alive():
                    move_responses = assoc.send_c_move(move_dataset, scp.ae_title,
                                                       query_model=C_MOVE_QUERY_MODEL)
                else:
                    raise Exception(f'Storage SCP failed to start for series {series_id}')

                check_responses(move_responses)

        dcm_path = os.path.join(self.dicom_dir, f'{instance_id}.dcm')
        png_path = process_and_write_png_from_file(dcm_path)
        return png_path

    def send_datasets(self, datasets: Iterable[Dataset], override_remote_ae: str = None,
                      override_pacs_url: str = None, override_pacs_port: int = None) -> None:
        if override_remote_ae is not None and override_pacs_url is not None and override_pacs_port is not None:
            send_remote_ae = override_remote_ae
            send_port = override_pacs_port
            send_url = override_pacs_url
        else:
            send_remote_ae = self.remote_ae
            send_port = self.pacs_port
            send_url = self.pacs_url

        ae = AE(ae_title=self.client_ae)
        ae.requested_contexts = StoragePresentationContexts
        with association(ae, send_url, send_port, send_remote_ae) as assoc:
            if assoc.is_established:
                for dataset in datasets:
                    logger.info('Sending %s', dataset.SeriesInstanceUID)
                    status = assoc.send_c_store(dataset)
                    for keyword in ['ErrorComment', 'OffendingElement']:
                        if getattr(status, keyword, None) is not None:
                            raise Exception(
                                f'failed to send because status[{keyword} = {getattr(status, keyword)}'
                            )
            else:
                raise Exception(
                    f'Unable to send because failed to establish association with {self.pacs_url}:{self.pacs_port}'
                )


def _find_patients(assoc, search_field, search_query, study_date_tag=None, additional_tags=None):
    dataset = Dataset()
    dataset.PatientID = None
    dataset.PatientName = ''
    dataset.PatientBirthDate = None
    if study_date_tag is not None:
        dataset.StudyDate = study_date_tag
    else:
        dataset.StudyDate = ''
    dataset.StudyInstanceUID = ''
    dataset.QueryRetrieveLevel = 'STUDY'
    setattr(dataset, search_field, search_query)
    set_undefined_tags_to_blank(dataset, additional_tags)
    return assoc.send_c_find(dataset, query_model=C_FIND_QUERY_MODEL)


socket_lock = threading.Lock()


class StorageSCP(threading.Thread):
    def __init__(self, client_ae, result_dir):
        self.result_dir = result_dir

        self.ae_title = f'{client_ae}'
        self.handlers = [(evt.EVT_C_STORE, self._on_c_store)]
        self.ae = AE(ae_title=self.ae_title)
        self.ae.supported_contexts = StoragePresentationContexts

        threading.Thread.__init__(self)

        self.daemon = True

    def run(self):
        """The thread run method"""
        self.ae.start_server(('0.0.0.0', 11113), evt_handlers=self.handlers)

    def stop(self):
        """Stop the SCP thread"""
        self.ae.shutdown()

    def path_for_dataset_instance(self, dataset):
        return os.path.join(self.result_dir, dicom_filename(dataset))

    def _on_c_store(self, event):
        '''
        :param dataset: pydicom.Dataset
            The DICOM dataset sent via the C-STORE
        :param context: pynetdicom3.presentation.PresentationContextTuple
            Details of the presentation context the dataset was sent under.
        :param info: dict
            A dict containing information about the association and DIMSE message.
        :return: pynetdicom.sop_class.Status or int
        '''
        dataset = event.dataset
        context = event.context
        try:
            logger.debug(f'_on_c_store called for {dataset.SOPInstanceUID}')

            os.makedirs(self.result_dir, exist_ok=True)
            filepath = self.path_for_dataset_instance(dataset)
            logger.info(f'Storing DICOM file: {filepath}')
            if os.path.exists(filepath):
                logger.warning('DICOM file already exists, overwriting')

            meta = Dataset()
            meta.MediaStorageSOPClassUID = dataset.SOPClassUID
            meta.MediaStorageSOPInstanceUID = dataset.SOPInstanceUID
            meta.ImplementationClassUID = PYNETDICOM_IMPLEMENTATION_UID
            meta.TransferSyntaxUID = context.transfer_syntax

            # The following is not mandatory, set for convenience
            meta.ImplementationVersionName = PYNETDICOM_IMPLEMENTATION_VERSION

            ds = FileDataset(filepath, {}, file_meta=meta, preamble=b"\0" * 128)
            ds.update(dataset)
            ds.is_little_endian = context.transfer_syntax.is_little_endian

            ds.is_implicit_VR = context.transfer_syntax.is_implicit_VR
            ds.save_as(filepath, write_like_original=False)

            status_ds = Dataset()
            status_ds.Status = 0x0000
        except Exception as e:
            logger.error(f'C-STORE failed: {e}')
            status_ds = Dataset()
            status_ds.Status = 0x0110  # Processing Failure
        return status_ds


@contextmanager
def association(ae, pacs_url, pacs_port, remote_aet, *args, **kwargs):
    try:
        assoc = ae.associate(pacs_url, pacs_port, ae_title=remote_aet, *args, **kwargs)
        if assoc.is_established:
            logger.debug(f'Association is established: {assoc}')
            yield assoc
        elif assoc.is_rejected:
            raise ConnectionError(f'Association rejected with {pacs_url}')
        elif assoc.is_aborted:
            raise ConnectionError(f'Received A-ABORT during association with {pacs_url}')
        else:
            raise ConnectionError(f'Failed to establish association with {pacs_url}')
    except Exception as e:
        raise e
    finally:
        logger.debug(f'Association being released: {assoc}')
        assoc.release()


@contextmanager
def storage_scp(client_ae, result_dir):
    try:
        scp = StorageSCP(client_ae, result_dir)
        socket_lock.acquire()
        scp.start()
        yield scp
    except Exception as e:
        raise e
    finally:
        scp.stop()
        scp.join()
        socket_lock.release()


def checked_responses(responses):
    '''
    Generator for checking success or pending status of DICOM responses
    Success response may only come once at the end of the dataset response list.

    :param responses: List of (Status, Dataset) tuples from pynetdicom call
    :return: List of Datasets or exception on warning/abort/failure
    '''
    for (status, dataset) in responses:
        logger.debug(status)
        logger.debug(dataset)
        if status.Status in status_success_or_pending:
            if isinstance(dataset, Dataset):
                yield dataset
        else:
            raise Exception('DICOM Response Failed With Status: 0x{0:04x}'.format(status.Status))


def check_responses(responses):
    for _ in checked_responses(responses):
        pass



from .dcmtk_client import _check_dcmtk_message_for_error


def test_stdout_error_checking():
    # Standard message, with error
    test_string = '''
D: ======================= END DIMSE MESSAGE =======================
I: Request Identifiers:
I: 
I: # Dicom-Data-Set
I: # Used TransferSyntax: Little Endian Implicit
I: (0008,0020) DA (no value available)                     #   0, 0 StudyDate
I: (0008,0052) CS [STUDY ]                                 #   6, 1 QueryRetrieveLevel
I: (0010,0010) PN (no value available)                     #   0, 0 PatientName
I: (0010,0020) LO [snipped]                                #  10, 1 PatientID
I: (0010,0030) DA (no value available)                     #   0, 0 PatientBirthDate
I: (0010,0040) CS (no value available)                     #   0, 0 PatientSex
I: (0020,000d) UI (no value available)                     #   0, 0 StudyInstanceUID
I: 
E: Find Failed, file: /tmp/tmp5ld6qp8a/find_input.dcm:
E: 0006:0207 DIMSE No data available (timeout in non-blocking mode)
E: Find SCU Failed: 0006:0207 DIMSE No data available (timeout in non-blocking mode)
I: Aborting Association''' # noqa
    assert(_check_dcmtk_message_for_error(test_string) == (0x0006, 0x0207))

    # Another error message
    test_string = '''
F: Association Request Failed: 0006:031b Failed to establish association
F: 0006:0317 Peer aborted Association (or never connected)
F: 0006:031c TCP Initialization Error: Connection refused'''
    assert(_check_dcmtk_message_for_error(test_string) == (0x006, 0x031c))

    # echoscu, no errors, verbose
    test_string = '''
I: Requesting Association
I: Association Accepted (Max Send PDV: 16372)
I: Sending Echo Request (MsgID 1)
I: Received Echo Response (Success)
I: Releasing Association
    '''
    assert(_check_dcmtk_message_for_error(test_string) is None)

    # Successful findscu, empty, verbose
    test_string = '''
I: Requesting Association
I: Association Accepted (Max Send PDV: 16372)
I: Sending Find Request (MsgID 1)
I: Request Identifiers:
I: 
I: # Dicom-Data-Set
I: # Used TransferSyntax: Little Endian Explicit
I: (0008,0052) CS [STUDY]                                  #   6, 1 QueryRetrieveLevel
I: (0008,103e) LO (no value available)                     #   0, 0 SeriesDescription
I: (0010,0010) PN [TEST]                                   #   4, 1 PatientName
I: 
I: Received Final Find Response (Success)
I: Releasing Association''' # noqa

    # Successful findscu, has results, non-verbose
    test_string = '''
I: ---------------------------
I: Find Response: 1 (Pending)
I: 
I: # Dicom-Data-Set
I: # Used TransferSyntax: Little Endian Explicit
I: (0008,0005) CS [ISO_IR 100]                             #  10, 1 SpecificCharacterSet
I: (0008,0052) CS [STUDY ]                                 #   6, 1 QueryRetrieveLevel
I: (0008,0054) AE [ORTHANC ]                               #   8, 1 RetrieveAETitle
I: (0008,103e) LO (no value available)                     #   0, 0 SeriesDescription
I: (0010,0010) PN [Lymphoma]                               #   8, 1 PatientName
I: 
I: ''' # noqa
    assert(_check_dcmtk_message_for_error(test_string) is None)

    # Handling an empty stdout
    assert(_check_dcmtk_message_for_error('') is None)



from .base_client import BaseDicomClient # noqa
from .pynetdicom_client import PynetDicomClient # noqa
from .dcmtk_client import DcmtkDicomClient # noqa
from .filesystem_dev_client import FilesystemDicomClient # noqa
from .utils import dataset_attribute_fetcher, copy_dicom_attributes, dicom_file_iterator # noqa



from abc import ABC, abstractmethod
from typing import List, Optional, Iterable

import pydicom
from pydicom import Dataset
from pydicom.valuerep import MultiValue
from pydicom.uid import UID

from .utils import getattr_required, copy_dicom_attributes


def _extend_datadict(datadict, tags):
    for tag in tags:
        try:
            existing_tag = datadict.get_entry(tag)
            if existing_tag != pacsman_private_tags[tag]:
                raise Exception(f'Private tag {tag} with different value already exists')
        except KeyError:
            pass
    datadict.add_private_dict_entries('pacsman', pacsman_private_tags)


# See this page in the DICOM standard for details on private tags:
# http://dicom.nema.org/medical/dicom/current/output/html/part05.html#sect_7.8
PRIVATE_ID = 'pacsman'

pacsman_private_tags = {
    0x00090010: ('LO', '1', 'Pacsman Private Identifier', '', 'PacsmanPrivateIdentifier'),
    0x00091001: ('UI', '1-N', 'Study Instance UIDs for Patient', '', 'PatientStudyInstanceUIDs'),
    0x00091002: ('DA', '1', 'Most Recent Study Date for Patient', '', 'PatientMostRecentStudyDate'),
}
_extend_datadict(pydicom.datadict, pacsman_private_tags)


class BaseDicomClient(ABC):
    @abstractmethod
    def verify(self) -> bool:
        """
        Send C-ECHO to PACS to verify connection
        :return: True on success, False on failure
        """
        raise NotImplementedError()

    @abstractmethod
    def search_patients(self, search_query: Optional[str] = None,
                        additional_tags: Optional[List[str]] = None,
                        wildcard: bool = True) -> List[Dataset]:
        """
        Search for patients. The PatientID and PatientName are searched.
        Performs a partial match if wildcard is True.
        :param search_query: String containing query value to c_find.
        :param additional_tags: additional DICOM tags for result datasets
        :param wildcard: Boolean stating whether to search partial matches (i.e. Sam would find Samuel).
            Defaults to True.
        :return: List of patient-Level pydicom Datasets, with tags:
            PatientName
            PatientID
            PatientBirthDate
            PatientStudyInstanceUIDs (private tag)
            PatientMostRecentStudyDate (private tag)
            Any valid DICOM tags in `additional_tags`
        """
        raise NotImplementedError()

    @abstractmethod
    def search_series(self, query_dataset, additional_tags=None) -> List[Dataset]:
        """
        Uses C-FIND to get patients matching the input (one req for id, one for name)
        :param query_dataset: Search dataset
        :param additional_tags: additional DICOM tags for result datasets
        :return: List of patient-Level pydicom Datasets, with tags:
            PatientName
            PatientID
            PatientBirthDate
            PatientStudyInstanceUIDs (private tag)
            PatientMostRecentStudyDate (private tag)
            Any valid DICOM tags in `additional_tags`
        """
        raise NotImplementedError()

    @abstractmethod
    def studies_for_patient(self, patient_id, study_date_tag=None, additional_tags=None) -> List[Dataset]:
        """
        Uses C-FIND to get study IDs for a patient.
        :param patient_id: Exact patient ID from PACS
        :param study_date_tag: Optional string in DICOM range format (e.g. '20200101-20200231') for studies
        :param additional_tags: additional DICOM tags for result datasets
        :return: List of pydicom Datasets with tags:
            PatientID
            StudyInstanceUID
            PatientName
            StudyDate
            Any valid DICOM tags in `additional_tags`
        """
        raise NotImplementedError()

    @abstractmethod
    def series_for_study(self, study_id, modality_filter=None, additional_tags=None,
                         manual_count=True) -> List[Dataset]:
        """
        :param study_id: StudyInstanceUID from PACS
        :param modality_filter: List of modalities to filter results on
        :param manual_count: if the PACS doesn't have the NumberOfSeriesRelatedInstances
          attribute, count it manually with an image level C-FIND per series.
        :param additional_tags: List of additioanl DICOM tags to add to result datasets
        :return: List of series-level pydicom Datasets, with tags:
            SeriesInstanceUID
            SeriesDescription
            SeriesDate
            SeriesTime
            Modality
            BodyPartExamined
            PatientPosition
            NumberOfSeriesRelatedInstances
            Any valid DICOM tags in `additional_tags`
        """
        raise NotImplementedError()

    def images_for_series(self, study_id, series_id, additional_tags=None, max_count=None) -> List[Dataset]:
        """
        :param study_id: StudyInstanceUID from PACS
        :param series_id: SeriesInstanceUID from PACS
        :param additional_tags:  List of additioanl DICOM tags to add to result datasets
        :param max_count: if not None then limits the number of images returned
        :return: list of image datasets
        """
        raise NotImplementedError()

    @abstractmethod
    def fetch_images_as_dicom_files(self, study_id: str, series_id: str) -> Optional[str]:
        """
        Fetches series images from PACS with C-MOVE
        :param study_id: StudyInstanceUID from PACS
        :param series_id: SeriesInstanceUID from PACS
        :return: a path to a directory full of dicom files on success, None if not found
        """
        raise NotImplementedError

    @abstractmethod
    def fetch_image_as_dicom_file(self, study_id: str, series_id: str, sop_instance_id: str) -> Optional[str]:
        """
        Fetches single series image from PACS with C-MOVE
        :param study_id: StudyInstanceUID from PACS
        :param series_id: SeriesInstanceUID from PACS
        :param sop_instance_id: SOPInstanceUID from PACS
        :return: a path to the dicom file on success, None if not found
        """
        raise NotImplementedError

    @abstractmethod
    def fetch_thumbnail(self, study_id: str, series_id: str) -> Optional[str]:
        """
        Fetches a central slice of a series from PACS and converts to PNG
        :param study_id: StudyInstanceUID from PACS
        :param series_id: SeriesInstanceUID from PACS
        :return: A path to a PNG file on success, None if not found
        """
        raise NotImplementedError

    @abstractmethod
    def fetch_slice_thumbnail(self, study_id: str, series_id: str,
                              instance_id: str) -> Optional[str]:
        """
        Fetches a specific slice of a series from PACS and converts to PNG
        :param study_id: StudyInstanceUID from PACS
        :param series_id: SeriesInstanceUID from PACS
        :param series_id: SeriesInstanceUID from PACS
        :param instance_id: SOPInstanceUID from PACS
        :return: A path to a PNG file on success, None if not found
        """
        raise NotImplementedError

    @staticmethod
    def update_patient_result(result, dataset, additional_tags=None):
        patient_id = getattr_required(dataset, 'PatientID')
        study_instance_uid = getattr_required(dataset, 'StudyInstanceUID')

        # Most of the data for a particular patient search result is grabbed
        # the first time this method is called for a patient.  This behaviour
        # may change in the future, e.g., we use the most recent attribute
        # values and/or combine attribute values from multiple different
        # datasets, when that data is missing in one or the other.
        # For now, we assume that the first time this method is called, the
        # `result` is an empty dataset.
        if len(result) == 0:
            result.PatientID = patient_id
            result.PatientName = getattr(dataset, 'PatientName', '')
            result.PatientBirthDate = getattr(dataset, 'PatientBirthDate', '')
            result.PatientStudyInstanceUIDs = MultiValue(UID, [study_instance_uid])
            result.PacsmanPrivateIdentifier = PRIVATE_ID
            result.PatientMostRecentStudyDate = getattr(dataset, 'StudyDate', '')
            copy_dicom_attributes(result, dataset, additional_tags, missing='empty')
        else:
            if result.PatientID != patient_id:
                raise ValueError("The search result has a different patient ID")

            existing_uids = {uid.name for uid in result.PatientStudyInstanceUIDs}
            if study_instance_uid.name not in existing_uids:
                result.PatientStudyInstanceUIDs.append(study_instance_uid)

        study_date = getattr(dataset, 'StudyDate', '')
        if study_date != '':
            no_existing_date = result.PatientMostRecentStudyDate == ''
            if no_existing_date or study_date > result.PatientMostRecentStudyDate:
                result.PatientMostRecentStudyDate = study_date

    @abstractmethod
    def send_datasets(self, datasets: Iterable[Dataset], override_remote_ae: str = None,
                      override_pacs_url: str = None, override_pacs_port: int = None) -> None:
        """
        Send a dicom dataset
        :param datasets: datasets to send
        :param override_remote_ae: override this clients stored AE and send to different remote
        :param override_pacs_url: conditionally required with override_remote_ae
        :param override_pacs_port: conditionally required with override_remote_ae
        :return:
        """
        raise NotImplementedError
